import { Injectable, LoggerService } from '@nestjs/common';
import { InjectConnection } from '@nestjs/typeorm';
import { Connection } from 'typeorm';
import { ConfigService } from '@nestjs/config';
import {
  Router,
  Gateway,
  GatewayGroup,
  RailgunRule,
  GatewayOption,
} from './config';
import * as IP from 'ip';
import * as _ from 'underscore';
import { AppLogger } from './logger.service';
import { User } from './entities/User';
import PQueue from 'p-queue';

const validDestinations = ['chnroute', 'all'];

export interface RouteDisplayData {
  id: number;
  name: string;
  description: string;
  type: 'gateway' | 'group';
  isLocal: boolean;
}

@Injectable()
export class AppService {
  gateways: Gateway[];
  routers: Router[];
  gatewayGroups: GatewayGroup[];
  railgunRules: RailgunRule[];
  syncQueue: PQueue;
  constructor(
    @InjectConnection('railgun')
    private db: Connection,
    private configService: ConfigService,
    private log: AppLogger,
  ) {
    this.log.setContext('Railgun');
    this.gateways = this.configService.get('gateways');
    this.routers = this.configService.get('routers');
    this.gatewayGroups = this.configService.get('gatewayGroups');
    this.railgunRules = this.configService.get('railgunRules');
    this.syncQueue = new PQueue({ concurrency: 1 });
    this.addSyncTask();
  }
  private async syncRouter(router: Router, users: User[]) {
    const ipsetData = this.getIpsetOfRouter(router, users);
    // replace this part with real ssh things
    console.log(router.name);
    console.log(ipsetData);
  }
  private async sync() {
    const users = await this.db.getRepository(User).find();
    await Promise.all(
      this.routers.map((router) => this.syncRouter(router, users)),
    );
  }
  private async addSyncTask() {
    return await this.syncQueue.add(async () => {
      this.sync();
    });
  }
  private subnetsContain(subnets: string[], ip) {
    return subnets.some((subnet) => IP.cidrSubnet(subnet).contains(ip));
  }
  private isGatewayAllowed(ip: string, gateway: Gateway, localOnly?: boolean) {
    if (gateway.hidden) {
      return false;
    }
    // 允许自己的网关的所有IP
    const router = gateway.routerInfo;
    if (router && this.subnetsContain(router.subnets, ip)) {
      return true;
    }
    if (localOnly) {
      return false;
    }
    // 在 railgun enterprise 表中有的IP
    return this.railgunRules.some((rule) => {
      return (
        _.contains(rule.gateways, `${gateway.router}/${gateway.isp}`) &&
        this.subnetsContain(rule.allowedIPs, ip)
      );
    });
  }
  private isGatewayGroupAllowed(ip: string, gatewayGroup: GatewayGroup) {
    // 在 railgun enterprise 表中有的IP
    return this.railgunRules.some((rule) => {
      return (
        _.contains(rule.groups, gatewayGroup.name) &&
        this.subnetsContain(rule.allowedIPs, ip)
      );
    });
  }
  private sortAddress(rawAddress: string) {
    if (IP.isV4Format(rawAddress)) {
      return rawAddress;
    }
    return rawAddress.slice(7);
  }
  private getAllSuitableGateways(ip: string) {
    return this.gateways.filter((gateway) =>
      this.isGatewayAllowed(ip, gateway, false),
    );
  }
  private getGatewayDisplayData(
    gateway: Gateway,
    ip: string,
  ): RouteDisplayData {
    return {
      id: gateway.id,
      name: `${gateway.router}/${gateway.isp}`,
      description: gateway.description,
      type: 'gateway',
      isLocal: this.subnetsContain(gateway.routerInfo.subnets, ip),
    };
  }
  private getGroupDisplayData(group: GatewayGroup): RouteDisplayData {
    return {
      id: group.id,
      name: group.name,
      description: group.description,
      type: 'group',
      isLocal: false,
    };
  }
  private async findOrCreateUser(ip: string) {
    const repo = this.db.getRepository(User);
    let user = await repo.findOne({
      where: { ip },
    });
    if (user) {
      return user;
    }
    user = new User();
    user.ip = ip;
    return await repo.save(user);
  }
  async getClientData(rawAddress: string) {
    const ip = this.sortAddress(rawAddress);
    const displayGateways = this.gateways
      .filter((gateway) => this.isGatewayAllowed(ip, gateway, false))
      .map((gateway) => this.getGatewayDisplayData(gateway, ip));
    const displayGroups = this.gatewayGroups
      .filter((group) => this.isGatewayGroupAllowed(ip, group))
      .map((group) => this.getGroupDisplayData(group));
    const user = await this.findOrCreateUser(ip);
    return {
      gateways: displayGateways.concat(displayGroups),
      user,
    };
  }

  private checkId(id: number, ip: string, localOnly: boolean) {
    if (!id) {
      return true;
    }
    if (id > 20000 && !localOnly) {
      const gatewayGroup = this.gatewayGroups.find((g) => g.id === id);
      return gatewayGroup && this.isGatewayGroupAllowed(ip, gatewayGroup);
    } else {
      const gateway = this.gateways.find((gw) => gw.id === id);
      return gateway && this.isGatewayAllowed(ip, gateway, localOnly);
    }
  }
  private getGatewayOrGroupFromId(id: number) {
    if (id > 20000) {
      const gatewayGroup = this.gatewayGroups.find((g) => g.id === id);
      return gatewayGroup;
    } else {
      const gateway = this.gateways.find((gw) => gw.id === id);
      return gateway;
    }
  }
  getIpsetNamesOfRouter(router: Router) {
    const ipsets: string[] = [];
    for (const destination of validDestinations) {
      const selfGateways = this.gateways.filter(
        (gateway) => router.name === gateway.router,
      );
      for (const gateway of selfGateways) {
        ipsets.push(`u_${gateway.isp}_${destination}`);
      }
      const otherRouters = this.routers.filter((r) => r !== router);
      for (const r of otherRouters) {
        ipsets.push(`u_${r.name.replace(/-/g, '-')}_${destination}`);
      }
      for (const group of this.gatewayGroups) {
        ipsets.push(`u_${group.name}_${destination}`);
      }
    }
    return ipsets;
  }
  getIpsetOfRouter(router: Router, users: User[]) {
    const lines: string[] = this.getIpsetNamesOfRouter(router).map(
      (setname) => `flush ${setname}`,
    );
    for (const group of this.gatewayGroups) {
      const matchUsers = users.filter(
        (user) =>
          user.remoteGatewayId === group.id &&
          this.isGatewayGroupAllowed(user.ip, group),
      );
      for (const user of matchUsers) {
        lines.push(`add u_${group.name.replace(/-/g, '_')}_${user.destination} ${user.ip}`);
      }
    }
    for (const gateway of this.gateways) {
      const matchUsers = users.filter(
        (user) =>
          user.remoteGatewayId === gateway.id &&
          this.isGatewayAllowed(user.ip, gateway, false),
      );
      for (const user of matchUsers) {
        if (gateway.routerInfo === router) {
          lines.push(`add u_${gateway.isp}_${user.destination} ${user.ip}`);
        } else {
          lines.push(
            `add u_${gateway.routerInfo.name.replace(/-/g, '_')}_${user.destination} ${user.ip}`,
          );
        }
      }
    }
    const selfGateways = this.gateways.filter(
      (gateway) => router.name === gateway.router,
    );
    for (const gateway of selfGateways) {
      const setName = `u_${gateway.isp}_all`;
      const matchUsers = users.filter(
        (user) =>
          user.localGatewayId === gateway.id &&
          this.isGatewayAllowed(user.ip, gateway, true),
      );
      for (const user of matchUsers) {
        lines.push(`add ${setName} ${user.ip}`);
      }
    }
    return lines.join('\n');
  }
  async postData(
    rawAddress: string,
    destination: string,
    remoteGatewayId: number,
    localGatewayId: number,
  ) {
    const ip = this.sortAddress(rawAddress);
    const user = await this.findOrCreateUser(ip);
    if (destination && !_.contains(validDestinations, destination)) {
      return {
        success: false,
        statusCode: 400,
        message: 'invalid destination',
      };
    }
    if (
      !this.checkId(remoteGatewayId, ip, false) ||
      !this.checkId(localGatewayId, ip, true)
    ) {
      return { success: false, statusCode: 404, message: 'invalid gateway' };
    }
    user.destination = destination || null;
    user.remoteGatewayId = remoteGatewayId || null;
    user.localGatewayId = localGatewayId || null;
    await this.db.getRepository(User).save(user);
    this.addSyncTask();
    return { success: true, statusCode: 200, message: 'success' };
  }
}
