import * as fs from 'fs';
import * as path from 'path';
import * as parse from 'csv-parse/lib/sync';

export interface Router {
  id: number;
  name: string;
  address: string;
  subnets: string[];
  lanInterfaces: string[];
  masqInterfaces: string[];
  location: string;
  user: string;
  host: string;
  sshPort: string;
  sshSystem: string;
  python: string;
  nextMark: number;
  destMark: number;
  port: number;
  offset: number;
  port2: number;
  offset2: number;
  wgPrivateKey: string;
  frpsPort: number;
}

export interface GatewayOption {
  id: number;
  description: string;
}

export interface Gateway extends GatewayOption {
  router: string;
  routerInfo: Router;
  isp: string;
  hidden: number;
  address: string;
  ipv4: string;
  ipv4Nat: string;
  ipv6: string;
  mtu: number;
  mark: number;
  bandwidth_inbound: number;
  bandwidth_outbound: number;
}

export interface RailgunRule {
  id: number;
  allowedIPs: string[];
  groups: string[];
  gateways: string[];
}

export interface GatewayGroup extends GatewayOption {
  name: string;
  locationPrefix: string[];
  includeRouters: string[];
  excludeRouters: string[];
  children: string[];
  destMark: number;
}

export interface Config {
  host: string;
  port: number;
  username: string;
  password: string;
  database: string;
  routers?: Router[];
  gateways?: Gateway[];
}

function loadSheet(sheetName: string, splitColumns: string[]) {
  const data = fs.readFileSync(
    path.join('data', `内网互联计划 - ${sheetName}.csv`),
  );

  const sheetData = parse(data, { columns: true, cast: true }).filter(
    (h) => h.id,
  );

  for (const columnName of splitColumns) {
    for (const data of sheetData) {
      const columnData = data[columnName];
      const splitedData = columnData ? columnData.split(',') : [];
      data[columnName] = splitedData;
    }
  }
  return sheetData;
}

export default function config() {
  const routers = loadSheet('nextgen2', [
    'subnets',
    'lanInterfaces',
    'splitInterfaces',
  ]) as Router[];
  const gateways = loadSheet('gateways2', []) as Gateway[];
  for (const gateway of gateways) {
    gateway.routerInfo = routers.find(
      (router) => router.name === gateway.router,
    );
  }
  return {
    db: {
      host: process.env.DB_HOST,
      port: process.env.DB_PORT ? parseInt(process.env.DB_PORT) : 5432,
      username: process.env.DB_USER,
      password: process.env.DB_PASS,
      database: process.env.DB_NAME,
    },
    routers,
    gateways,
    gatewayGroups: loadSheet('gateway groups', [
      'locationPrefix',
      'includeRouters',
      'excludeRouters',
      'children',
    ]) as GatewayGroup[],
    railgunRules: loadSheet('railgun enterprise', [
      'allowedIPs',
      'groups',
      'gateways',
    ]) as RailgunRule[],
  };
}
